#ifndef CFMUVALIDATESERVER_H
#define CFMUVALIDATESERVER_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "sysdepsgui.h"

#include <gtkmm.h>
#include <sigc++/sigc++.h>
#ifdef HAVE_WEBKITGTK
#include <webkit/webkit.h>
#endif

#include "tfr.hh"

class CFMUValidateServer : public sigc::trackable {
public:
	CFMUValidateServer(int& argc, char**& argv);
	~CFMUValidateServer();
	void run(void);

protected:
	AirportsDb m_airportdb;
	NavaidsDb m_navaiddb;
	WaypointsDb m_waypointdb;
	AirwaysDb m_airwaydb;
	AirspacesDb m_airspacedb;
	Engine *m_engine;
	TrafficFlowRestrictions m_tfr;
	std::string m_dir_main;
	std::string m_dir_aux;
        Glib::RefPtr<Glib::MainLoop> m_mainloop;
        Glib::RefPtr<Glib::IOChannel> m_inputchan;
        Glib::RefPtr<Glib::IOChannel> m_outputchan;
        sigc::connection m_inputreaderconn;
        sigc::connection m_delayedstatechangeconn;
	std::list<std::string> m_flightplans;
	GtkWidget *m_portal_webview;
	GtkWidget *m_validate_webview;
	gulong m_portal_webview_close_sighandler;
	gulong m_portal_webview_create_sighandler;
	gulong m_portal_webview_ready_sighandler;
	gulong m_portal_webview_loadstatus_sighandler;
	gulong m_portal_webview_navpolicy_sighandler;
	gulong m_portal_webview_newwndpolicy_sighandler;
	gulong m_portal_webview_rsrcload_sighandler;
	gulong m_portal_webview_consolemsg_sighandler;
	gulong m_validate_webview_close_sighandler;
	gulong m_validate_webview_create_sighandler;
	gulong m_validate_webview_ready_sighandler;
	gulong m_validate_webview_loadstatus_sighandler;
	gulong m_validate_webview_navpolicy_sighandler;
	gulong m_validate_webview_newwndpolicy_sighandler;
	gulong m_validate_webview_rsrcload_sighandler;
	gulong m_validate_webview_consolemsg_sighandler;
	guint m_verbose;
	typedef enum {
		wvstate_loadmainpage,
		wvstate_loadquerypage,
		wvstate_ready,
		wvstate_query,
		wvstate_waitresult,
		wvstate_failed
	} wvstate_t;
	wvstate_t m_wvstate;
	Engine::auxdb_mode_t m_auxdbmode;
	bool m_quit;
	bool m_local;
	bool m_tfrloaded;

	bool input_line_handler(Glib::IOCondition iocond);
	void handle_input(void);
	void print(const TrafficFlowRestrictions::Message& m);
	void print(const TrafficFlowRestrictions::Result& res);
	void print(const TrafficFlowRestrictions::RuleResult& rule);
	void print_tracerules(void);
	void print_disabledrules(void);
	void add_portal_sighandlers(void);
	void remove_portal_sighandlers(void);
	void add_validate_sighandlers(void);
	void remove_validate_sighandlers(void);
	void prepare_webview(void);
	void set_wvstate(wvstate_t newstate);
	void parse_results(void);
	void zap_errors(void);
	static gboolean close_webview_1(WebKitWebView* webview, gpointer userdata);
	gboolean close_webview(WebKitWebView* webview);
	static WebKitWebView *create_webview_1(WebKitWebView *webview, WebKitWebFrame *frame, gpointer userdata);
	WebKitWebView *create_webview(WebKitWebView *webview, WebKitWebFrame *frame);
	static gboolean webview_ready_1(WebKitWebView* webview, gpointer userdata);
	gboolean webview_ready(WebKitWebView* webview);
	static void loadstatus_webview_1(WebKitWebView *webview, GParamSpec *pspec, gpointer userdata);
	void loadstatus_webview(WebKitWebView *webview, GParamSpec *pspec);
	static gboolean policy_decision_required_1(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitNetworkRequest *request,
						   WebKitWebNavigationAction *action, WebKitWebPolicyDecision *decision, gpointer userdata);
	gboolean policy_decision_required(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitNetworkRequest *request,
					  WebKitWebNavigationAction *action, WebKitWebPolicyDecision *decision);
	static void resourceload_finished_1(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitWebResource *webresource, gpointer userdata);
	void resourceload_finished(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitWebResource *webresource);
	static gboolean result_change_1(WebKitDOMElement *target, WebKitDOMEvent *e, gpointer userdata);
	gboolean result_change(WebKitDOMElement *target, WebKitDOMEvent *e);
	static gboolean console_message_1(WebKitWebView *webview, gchar *message, gint line, gchar *source_id, gpointer userdata);
	gboolean console_message(WebKitWebView *webview, gchar *message, gint line, gchar *source_id);
};

#endif /* CFMUVALIDATESERVER_H */
