#ifndef INTERVAL_H
#define INTERVAL_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <limits>
#include <string>
#include <set>

template <typename T>
class IntervalSet {
public:
	typedef T type_t;

	class Interval {
	public:
		Interval(void);
		Interval(type_t val);
		Interval(type_t val0, type_t val1);

		type_t get_lower(void) const { return m_lower; }
		type_t get_upper(void) const { return m_upper; }
		bool operator<(const Interval& x) const;
		bool is_empty(void) const;
		bool is_full(void) const;
		bool is_inside(type_t x) const;
		bool is_overlap(const Interval& x) const;
		bool is_adjacent(const Interval& x) const;
		bool is_overlap_or_adjacent(const Interval& x) const;
		void set_empty(void);
		void set_full(void);
		std::string to_str(void) const;

	protected:
		type_t m_lower;
		type_t m_upper;
	};

	IntervalSet(void);
	IntervalSet(const Interval& x);
	IntervalSet(const IntervalSet& x);
	bool is_empty(void) const;
	bool is_full(void) const;
	bool is_inside(type_t x) const;
	void set_empty(void);
	void set_full(void);
	IntervalSet& operator&=(type_t x);
	IntervalSet& operator|=(type_t x);
	IntervalSet& operator^=(type_t x);
	IntervalSet& operator-=(type_t x);
	IntervalSet& operator&=(const IntervalSet::Interval& x);
	IntervalSet& operator|=(const IntervalSet::Interval& x);
	IntervalSet& operator^=(const IntervalSet::Interval& x);
	IntervalSet& operator-=(const IntervalSet::Interval& x);
	IntervalSet& operator&=(const IntervalSet& x);
	IntervalSet& operator|=(const IntervalSet& x);
	IntervalSet& operator^=(const IntervalSet& x);
	IntervalSet& operator-=(const IntervalSet& x);
	IntervalSet operator~() const;
	std::string to_str(void) const;

protected:
	typedef std::set<Interval> set_t;

public:
	typedef typename set_t::const_iterator const_iterator;
	const_iterator begin(void) const { return m_set.begin(); }
	const_iterator end(void) const { return m_set.end(); }

protected:
	set_t m_set;

	void normalize(void);
};

template <typename T> IntervalSet<T> operator&(const IntervalSet<T>& x, const IntervalSet<T>& y)
{
	IntervalSet<T> r(x);
	r &= y;
	return r;
}

template <typename T> IntervalSet<T> operator|(const IntervalSet<T>& x, const IntervalSet<T>& y)
{
	IntervalSet<T> r(x);
	r |= y;
	return r;
}

template <typename T> IntervalSet<T> operator^(const IntervalSet<T>& x, const IntervalSet<T>& y)
{
	IntervalSet<T> r(x);
	r ^= y;
	return r;
}

template <typename T> IntervalSet<T> operator-(const IntervalSet<T>& x, const IntervalSet<T>& y)
{
	IntervalSet<T> r(x);
	r -= y;
	return r;
}

#endif /* INTERVAL_H */
