/***************************************************************************
 *   Copyright (C) 2007, 2008, 2009, 2011, 2012 by Thomas Sailer           *
 *   t.sailer@alumni.ethz.ch                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "sysdeps.h"

#include <iomanip>

#include "flightdeckwindow.h"
#include "SunriseSunset.h"

FlightDeckWindow::SunriseSunsetWaypointListModelColumns::SunriseSunsetWaypointListModelColumns(void)
{
	add(m_icao);
	add(m_name);
	add(m_sunrise);
	add(m_sunset);
	add(m_lon);
	add(m_lat);
	add(m_mt);
	add(m_tt);
	add(m_dist);
	add(m_point);
	add(m_nr);
}

void FlightDeckWindow::srsspage_recompute(void)
{
	if (!m_srssstore)
		return;
	if (!m_srsscalendar) {
		for (Gtk::TreeIter iter = m_srssstore->children().begin(); iter; iter++) {
			Gtk::TreeModel::Row row(*iter);
			row[m_srsscolumns.m_sunrise] = "";
			row[m_srsscolumns.m_sunset] = "";
		}
		return;
	}
	Glib::Date dt;
	m_srsscalendar->get_date(dt);
        for (Gtk::TreeIter iter = m_srssstore->children().begin(); iter; iter++) {
                Gtk::TreeModel::Row row(*iter);
                double sr, ss;
                int r = SunriseSunset::civil_twilight(dt.get_year(), dt.get_month(), dt.get_day(), row[m_srsscolumns.m_point], sr, ss);
                if (r) {
                        row[m_srsscolumns.m_sunrise] = _("always");
                        row[m_srsscolumns.m_sunset] = (r < 0) ? _("night") : _("day");
                        continue;
                }
                if (sr < 0.0)
                        sr += 24.0;
                if (ss < 0.0)
                        ss += 24.0;
                int h = (int)sr;
                int m = (int)((sr - h) * 60 + 0.5);
                char buf[8];
                snprintf(buf, sizeof(buf), "%02d:%02d", h, m);
                row[m_srsscolumns.m_sunrise] = buf;
                h = (int)ss;
                m = (int)((ss - h) * 60 + 0.5);
                snprintf(buf, sizeof(buf), "%02d:%02d", h, m);
                row[m_srsscolumns.m_sunset] = buf;
        }
}

void FlightDeckWindow::srsspage_updatemenu(void)
{
	if (m_srssstore)
		m_srssstore->clear();
	if (m_menuid != 303)
		return;
	if (m_srssstore && m_sensors) {
		const FPlanRoute& fpl(m_sensors->get_route());
		for (unsigned int nr = 0; nr < fpl.get_nrwpt(); ++nr) {
			Gtk::TreeModel::Row row(*(m_srssstore->append()));
			const FPlanWaypoint& wpt(fpl[nr]);
			row[m_srsscolumns.m_icao] = wpt.get_icao();
			row[m_srsscolumns.m_name] = wpt.get_name();
			row[m_srsscolumns.m_lon] = wpt.get_coord().get_lon_str();
			row[m_srsscolumns.m_lat] = wpt.get_coord().get_lat_str();
			if (nr + 1 < fpl.get_nrwpt()) {
				const FPlanWaypoint& wpt2(fpl[nr + 1]);
				Point pt1(wpt.get_coord()), pt2(wpt2.get_coord());
				float dist = pt1.spheric_distance_nmi(pt2);
				float tt = pt1.spheric_true_course(pt2);
				WMM wmm;
				wmm.compute((wpt.get_altitude() + wpt2.get_altitude()) * (FPlanWaypoint::ft_to_m / 2000.0), pt1.halfway(pt2), time(0));
				float mt = tt;
				if (wmm)
					mt -= wmm.get_dec();
				row[m_srsscolumns.m_mt] = Conversions::track_str(mt);
				row[m_srsscolumns.m_tt] = Conversions::track_str(tt);
				row[m_srsscolumns.m_dist] = Conversions::dist_str(dist);;
			} else {
				row[m_srsscolumns.m_mt] = "";
				row[m_srsscolumns.m_tt] = "";
				row[m_srsscolumns.m_dist] = "";
			}
			row[m_srsscolumns.m_point] = wpt.get_coord();
			row[m_srsscolumns.m_nr] = nr + 1;
		}
	}
	if (m_srsscalendar) {
		m_srsscalendar->clear_marks();
		Glib::Date today;
		today.set_time_current();
		m_srsscalendar->select_month(today.get_month() - 1, today.get_year());
		m_srsscalendar->select_day(today.get_day());
		m_srsscalendar->mark_day(today.get_day());
	}
	srsspage_recompute();
}
