//
// C++ Interface: airdata
//
// Description: AirData related routines
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2012
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef AIRDATA_H
#define AIRDATA_H

#include "baro.h"

class AirData {
  public:
	static const float gamma = 1.402;
	static const float speed_of_sound_0degc = 331.3 * 3.6 / 1.852 /*Point::mps_to_kts*/;
	static const float speed_of_sound_std;

	// units of measurement:
	// altitude: ft
	// temperature: degrees centigrade
	// pressure: hPa
	// speed: kts

	AirData(void);

	void set_qnh(float qnh = IcaoAtmosphere::std_sealevel_pressure);
	void set_qnh_temp(float qnh = IcaoAtmosphere::std_sealevel_pressure, float temp = IcaoAtmosphere::std_sealevel_temperature);
	void set_qnh_tempoffs(float qnh = IcaoAtmosphere::std_sealevel_pressure, float tempoffs = 0) { set_qnh_temp(qnh, tempoffs + IcaoAtmosphere::std_sealevel_temperature); }
	float get_qnh(void) const { return m_atmosphere.get_qnh(); }
	float get_temp(void) const { return m_atmosphere.get_temp(); }
	float get_tempoffs(void) const { return get_temp() - IcaoAtmosphere::std_sealevel_temperature; }

	const IcaoAtmosphere& get_atmosphere(void) const { return m_atmosphere; }

	float true_to_pressure_altitude(float truealt) const;
	float pressure_to_true_altitude(float baroalt) const;

	static void std_altitude_to_pressure(float *press, float *temp, float alt) { IcaoAtmosphere::std_altitude_to_pressure(press, temp, alt); }
	static void std_pressure_to_altitude(float *alt, float *temp, float press) { IcaoAtmosphere::std_pressure_to_altitude(alt, temp, press); }
	void altitude_to_pressure(float *press, float *temp, float alt) const { m_atmosphere.altitude_to_pressure(press, temp, alt); }
	void pressure_to_altitude(float *alt, float *temp, float press) const { m_atmosphere.pressure_to_altitude(alt, temp, press); }

	float get_pressure_altitude(void) const { return m_pressalt; }
	void set_pressure_altitude(float pressalt);
	float get_true_altitude(void) const { return m_truealt; }
	void set_true_altitude(float truealt);
	float get_density_altitude(void) const;
	void set_density_altitude(float da);

	float get_tatprobe_ct(void) const { return m_ct; }
	void set_tatprobe_ct(float ct = 1);

	float get_oat(void) const { return m_temp - IcaoAtmosphere::degc_to_kelvin; }
	void set_oat(float oat);
	float get_rat(void) const;
	void set_rat(float oat);

	float get_cas(void) const { return m_cas; }
	float get_eas(void) const { return m_eas; }
	float get_tas(void) const { return m_tas; }
	float get_mach(void) const { return m_mach; }
	void set_cas(float cas);
	void set_tas(float tas);

	float get_p(void) const { return m_p; }
	float get_q(void) const { return m_q; }
	static float get_lss(float temp);
	float get_lss(void) const { return get_lss(get_oat()); }

  protected:
	IcaoAtmosphere m_atmosphere;
	float m_pressalt;
	float m_truealt;
	mutable float m_densityalt;
	float m_p;
	float m_q;
	float m_cas;
	float m_eas;
	float m_tas;
	float m_mach;
	float m_ct;
	float m_temp;
	float m_tempstd;

	float get_tat_sat_ratio(void) const;
	void compute_mach_eas_tas(void);
	void compute_tas(void);
	void compute_cas(void);
};

#endif /* AIRDATA_H */
