//
// C++ Interface: icaofpl
//
// Description: 
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2012
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef ICAOFPL_H
#define ICAOFPL_H

#include <string>
#include <map>
#include "sysdeps.h"
#include "engine.h"
#include "fplan.h"
#include "aircraft.h"

class IcaoFlightPlan {
  public:
	typedef enum {
		awymode_keep,
		awymode_collapse,
		awymode_collapse_all
	} awymode_t;

	IcaoFlightPlan(Engine& engine);
	std::pair<std::string::const_iterator,std::string> parse(std::string::const_iterator txti, std::string::const_iterator txte, bool expand_airways = true);
	std::string parse(const std::string& txt, bool expand_airways = true) { return parse(txt.begin(), txt.end(), expand_airways).second; }
	void populate(const FPlanRoute& route, awymode_t awymode = awymode_collapse, double dct_limit = 50.0);
	std::string get_fpl(void);
	void set_route(FPlanRoute& route);
	void recompute_eet(void);

	const std::string& get_aircraftid(void) const { return m_aircraftid; }
	char get_flightrules(void) const;
	char get_flighttype(void) const { return m_flighttype; }
	unsigned int get_number(void) const { return m_number; }
	const std::string& get_aircrafttype(void) const { return m_aircrafttype; }
	char get_wakecategory(void) const { return m_wakecategory; }
	const std::string& get_equipment(void) const { return m_equipment; }
	const std::string& get_transponder(void) const { return m_transponder; }
	const std::string& get_pbn(void) const { return m_pbn; }
	const std::string& get_cruisespeed(void) const { return m_cruisespeed; }
	const std::string& get_departure(void) const { return m_departure; }
	const std::string& get_destination(void) const { return m_destination; }
	const std::string& get_alternate1(void) const { return m_alternate1; }
	const std::string& get_alternate2(void) const { return m_alternate2; }
	time_t get_departuretime(void) const { return m_departuretime; }
	time_t get_totaleet(void) const { return m_totaleet; }
	time_t get_endurance(void) const { return m_endurance; }
	unsigned int get_personsonboard(void) const { return m_personsonboard; }
	const std::string& get_colourmarkings(void) const { return m_colourmarkings; }
	const std::string& get_remarks(void) const { return m_remarks; }
	const std::string& get_picname(void) const { return m_picname; }
	const std::string& get_emergencyradio(void) const { return m_emergencyradio; }
	const std::string& get_survival(void) const { return m_survival; }
	const std::string& get_lifejackets(void) const { return m_lifejackets; }
	const std::string& get_dinghies(void) const { return m_dinghies; }

	void set_aircraft(const Aircraft& acft,
				 double rpm = std::numeric_limits<double>::quiet_NaN(),
				 double mp = std::numeric_limits<double>::quiet_NaN(),
				 double bhp = std::numeric_limits<double>::quiet_NaN());

	void set_aircraftid(const std::string& x) { m_aircraftid = x; }
	void set_flighttype(char x) { m_flighttype = x; }
	void set_number(unsigned int x) { m_number = x; }
	void set_aircrafttype(const std::string& x) { m_aircrafttype = x; }
	void set_wakecategory(char x) { m_wakecategory = x; }
	void set_equipment(const std::string& x) { m_equipment = x; }
	void set_transponder(const std::string& x) { m_transponder = x; }
	void set_pbn(const std::string& x) { m_pbn = x; }
	void set_cruisespeed(const std::string& x) { m_cruisespeed = x; }
	void set_departure(const std::string& x) { m_departure = x; }
	void set_destination(const std::string& x) { m_destination = x; }
	void set_alternate1(const std::string& x) { m_alternate1 = x; }
	void set_alternate2(const std::string& x) { m_alternate2 = x; }
	void set_departuretime(time_t x);
	void set_totaleet(time_t x) { m_totaleet = x; }
	void set_endurance(time_t x) { m_endurance = x; }
	void set_personsonboard(unsigned int x) { m_personsonboard = x; }
	void set_personsonboard_tbn(void) { m_personsonboard = ~0U; }
	void set_colourmarkings(const std::string& x) { m_colourmarkings = x; }
	void set_remarks(const std::string& x) { m_remarks = x; }
	void set_picname(const std::string& x) { m_picname = x; }
	void set_emergencyradio(const std::string& x) { m_emergencyradio = x; }
	void set_survival(const std::string& x) { m_survival = x; }
	void set_lifejackets(const std::string& x) { m_lifejackets = x; }
	void set_dinghies(const std::string& x) { m_dinghies = x; }

	void otherinfo_clear(void) { m_otherinfo.clear(); }
	void otherinfo_clear(const std::string& category);
	const std::string& otherinfo_get(const std::string& category);
	void otherinfo_add(const Aircraft::OtherInfo& oi) { otherinfo_add(oi.get_category(), oi.get_text()); }
	void otherinfo_add(const std::string& category, const std::string& text) { otherinfo_add(category, text.begin(), text.end()); }
	void otherinfo_add(const std::string& category, std::string::const_iterator texti, std::string::const_iterator texte);
	static std::vector<std::string> tokenize(const std::string& s);
	static std::string untokenize(const std::vector<std::string>& s);

	bool mark_vfrroute_begin(FPlanRoute& route);
	bool mark_vfrroute_end(FPlanRoute& route);

	static bool is_aiport_paris_tma(const std::string& icao); // Paris group airport
	static bool is_aiport_lfpnv(const std::string& icao);
	static bool is_aiport_lfob(const std::string& icao);
	static bool is_route_pogo(const std::string& dep, const std::string& dest); // POGO airport group
	bool is_route_pogo(void) const { return is_route_pogo(get_departure(), get_destination()); }

	class FindCoord {
	  public:
		static const unsigned int flag_airport    = 1 << 0;
		static const unsigned int flag_navaid     = 1 << 1;
		static const unsigned int flag_waypoint   = 1 << 2;
		static const unsigned int flag_mapelement = 1 << 3;
		static const unsigned int flag_coord      = 1 << 4;
		static const unsigned int flag_subtables  = 1 << 31;

		FindCoord(Engine& engine);
		bool find(const std::string& icao, const std::string& name, unsigned int flags, const Point& pt = Point());
		bool find(const Point& pt, unsigned int flags, float maxdist_km);
		bool find_airway(const std::string& awy);
		bool find_area(const Rect& rect, Engine::AirwayGraphResult::Vertex::typemask_t tmask = Engine::AirwayGraphResult::Vertex::typemask_all);
		std::string get_name(void) const;
		std::string get_icao(void) const;
		Point get_coord(void) const;
		bool get_airport(AirportsDb::element_t& x) const;
		bool get_navaid(NavaidsDb::element_t& x) const;
		bool get_waypoint(WaypointsDb::element_t& x) const;
		bool get_mapelement(MapelementsDb::element_t& x) const;
		bool get_coord(Point& x) const;
		bool get_airwaygraph(Engine::AirwayGraphResult::Graph& x) const;

	  protected:
		class SortAirportNamelen;
		class SortNavaidNamelen;
		class SortMapelementNamelen;
		class SortAirportDist;
		class SortNavaidDist;
		class SortWaypointDist;
		class SortMapelementDist;

		static const std::string empty;

		Engine::AirportResult::elementvector_t m_airports;
		Engine::NavaidResult::elementvector_t m_navaids;
		Engine::WaypointResult::elementvector_t m_waypoints;
		Engine::MapelementResult::elementvector_t m_mapelements;
		Engine::AirwayGraphResult::Graph m_airwaygraph;
		Glib::RefPtr<Engine::AirportResult> m_airportquery;
		Glib::RefPtr<Engine::AirportResult> m_airportquery1;
		Glib::RefPtr<Engine::NavaidResult> m_navaidquery;
		Glib::RefPtr<Engine::NavaidResult> m_navaidquery1;
		Glib::RefPtr<Engine::WaypointResult> m_waypointquery;
		Glib::RefPtr<Engine::MapelementResult> m_mapelementquery;
		Glib::RefPtr<Engine::AirwayGraphResult> m_airwayquery;
		Glib::RefPtr<Engine::AreaGraphResult> m_areaquery;
		Glib::Cond m_cond;
		Glib::Mutex m_mutex;
		Engine& m_engine;
		Point m_coordresult;

		void async_cancel(void);
		void async_clear(void);
		void async_connectdone(void);
		void async_done(void);
		unsigned int async_waitdone(void);
		void sort_dist(const Point& pt);
		void cut_dist(const Point& pt, float maxdist_km);
		void retain_shortest_dist(const Point& pt);
	};

  protected:
	class Waypoint {
	  public:
                Waypoint(void);
                Waypoint(const FPlanWaypoint& wpt);

                const std::string& get_icao(void) const { return m_icao; }
                void set_icao(const std::string& t) { m_icao = t; }
                const std::string& get_name(void) const { return m_name; }
                void set_name(const std::string& t) { m_name = t; }
		FPlanWaypoint::pathcode_t get_pathcode(void) const { return m_pathcode; }
		void set_pathcode(FPlanWaypoint::pathcode_t pc) { m_pathcode = pc; }
                const std::string& get_pathname(void) const { return m_pathname; }
                void set_pathname(const std::string& t) { m_pathname = t; }
		const std::string& get_icao_or_name(void) const;
                time_t get_time(void) const { return m_time; }
                void set_time(time_t t) { m_time = t; }
                const Point& get_coord(void) const { return m_coord; }
                Point::coord_t get_lon(void) const { return m_coord.get_lon(); }
                Point::coord_t get_lat(void) const { return m_coord.get_lat(); }
                void set_coord(const Point& c) { m_coord = c; }
                void set_lon(Point::coord_t lo) { m_coord.set_lon(lo); }
                void set_lat(Point::coord_t la) { m_coord.set_lat(la); }
                void set_lon_deg(float lo) { m_coord.set_lon_deg(lo); }
                void set_lat_deg(float la) { m_coord.set_lat_deg(la); }
		std::string get_coordstr(void) const;
                int16_t get_altitude(void) const { return m_alt; }
                void set_altitude(int16_t a) { m_alt = a; }
                uint16_t get_flags(void) const { return m_flags; }
                void set_flags(uint16_t f) { m_flags = f; }
                uint32_t get_frequency(void) const { return m_frequency; }
                void set_frequency(uint32_t f) { m_frequency = f; }
		bool enforce_pathcode_vfrifr(void);

        private:
                std::string m_icao;
                std::string m_name;
                std::string m_pathname;
                time_t m_time;
                Point m_coord;
                int16_t m_alt;
                uint16_t m_flags;
		uint32_t m_frequency;
		FPlanWaypoint::pathcode_t m_pathcode;
	};
	
	Engine& m_engine;
	typedef std::vector<Waypoint> route_t;
	route_t m_route;
	typedef std::map<std::string,std::string> otherinfo_t;
	otherinfo_t m_otherinfo;
	std::string m_aircraftid;
	std::string m_aircrafttype;
	std::string m_equipment;
	std::string m_transponder;
	std::string m_pbn;
	std::string m_cruisespeed;
	std::string m_departure;
	std::string m_destination;
	std::string m_alternate1;
	std::string m_alternate2;
	std::string m_sid;
	std::string m_star;
	std::string m_colourmarkings;
	std::string m_remarks;
	std::string m_picname;
	std::string m_emergencyradio;
	std::string m_survival;
	std::string m_lifejackets;
	std::string m_dinghies;
	typedef std::map<int,float> cruisespeeds_t;
	cruisespeeds_t m_cruisespeeds;
	time_t m_departuretime;
	time_t m_totaleet;
	time_t m_endurance;
	unsigned int m_number;
	unsigned int m_personsonboard;
	uint16_t m_departureflags;
	uint16_t m_destinationflags;
	unsigned int m_defaultalt;
	char m_flighttype;
	char m_wakecategory;

	Rect get_bbox(void) const;
	typedef std::set<std::string> nameset_t;
	static nameset_t intersect(const nameset_t& s1, const nameset_t& s2);
	float get_cruisespeed(int& alt) const;
	bool has_pathcodes(void) const;
	bool enforce_pathcode_vfrdeparr(void);
	bool enforce_pathcode_vfrifr(void);
	void erase_vfrroute_begin(void);
	void erase_vfrroute_end(void);
	void redo_route_names(void);
	void find_airways(void);
	void fix_max_dct_distance(double dct_limit = 50.0);
	void erase_unnecessary_airway(bool keep_turnpoints = true);
	void add_eet(void);
	void normalize_pogo(void);
	std::string expand_airways(bool really_expand = true);
	static void trimleft(std::string::const_iterator& txti, std::string::const_iterator txte);
	static void trimright(std::string::const_iterator txti, std::string::const_iterator& txte);
	static void trim(std::string::const_iterator& txti, std::string::const_iterator& txte);
	static std::string upcase(const std::string& txt);
	static bool parsetxt(std::string& txt, unsigned int len, std::string::const_iterator& txti, std::string::const_iterator txte);
	static bool parsenum(unsigned int& num, unsigned int digits, std::string::const_iterator& txti, std::string::const_iterator txte);
	static bool parsetime(time_t& t, std::string::const_iterator& txti, std::string::const_iterator txte);
	static bool parsespeed(double& spd, std::string::const_iterator& txti, std::string::const_iterator txte);
	static bool parsespeed(std::string& spd, std::string::const_iterator& txti, std::string::const_iterator txte);
	static bool parsealt(int& alt, unsigned int& flags, std::string::const_iterator& txti, std::string::const_iterator txte);

	static const unsigned int parsewpt_coords   = 1 << 0;
	static const unsigned int parsewpt_airway   = 1 << 1;
	static const unsigned int parsewpt_frules   = 1 << 2;
	static const unsigned int parsewpt_speedalt = 1 << 3;
	unsigned int parsewpt(Waypoint& wpt, std::string::const_iterator& txti, std::string::const_iterator txte);
};

#endif /* ICAOFPL_H */
